#library(igraph)
library(glasso)
library(MASS)
library(clues)
library(igraph)#, lib.loc="/share/apps/R/packages/")
library(huge)#, lib.loc="/share/apps/R/packages/")
#library(SGL)

GeneratorBlock <- function(n_sample=200, d=128, n_block=4, prob_in=0.2, prob_out=0.02)
{
  average_size <- d / n_block;
  size <- as.integer( rnorm(n_block-1, average_size, average_size/10) )
  size <- c(size, d - sum(size) )
  if( any(size <= 0) )
  {
    cat( "para set error")
    return()
  }
  
  node_cluster <- rep(0, d)
  index <- 1
  for(i in 1:length(size) )
    for(j in 1:size[i])
    {
      node_cluster[index] <- i
      index <- index + 1
    }
  
  omega <- matrix(0, d, d)
  for(i in 1:(d-1) )
  {
    for(j in (i+1):d)
    {
      p <- runif(1, 0, 1)
      if(node_cluster[i] == node_cluster[j] && p <= prob_in)
      {
        omega[i,j] <- 0.3
        omega[j,i] <- omega[i,j]
      }
      if(node_cluster[i] != node_cluster[j] && p <= prob_out)
      {
        omega[i,j] <- 0.3
        omega[j,i] <- omega[i,j]
      }
    }
  }
  diag(omega) <- abs( min(eigen(omega)$values) ) + 0.2
  
  #write("start\tend", "tmp_mat")
  #for(i in 1:(d-1) )
   # for(j in (i+1):d)
    #  if(omega[i,j] > 0.00001)
     #   write(c(i,j), "tmp_mat", append=T, sep="\t")
  sigma <- solve(omega)
  data <- mvrnorm(n_sample, rep(0, d), sigma)
  return( list(data=data, omega=omega, sigma=sigma, sigma_hat=cov(data),cluster=node_cluster) )
}

VisualMatrix <- function(m) {
	m <- abs(m)
  	tt <- t(m)[,nrow(m):1] 
	image(tt, col = grey(seq(1, 0, length = 256)))
}

CompareCluster <- function(true.cluster, pred.cluster)
	#true.cluster <- L$cluster
	#pred.cluster <- pred.res$membership
{
	return (adjustedRand(true.cluster, pred.cluster)[1] )	
}

PathAcc <- function(true.cluster, path)
{
	n <- length(path[1,])
	acc <- rep(0, n)
	for(i in 1:n)
		acc[i] <- CompareCluster(true.cluster, path[,i])
	return(acc)
}

PressMatrix <- function(mat, cluster) 
{
	n <- length(mat[1,])
	for(i in 1:(n-1))
		for(j in (i+1):n)
		{
			#cat(i, j, "\n")
    		if( cluster[i] != cluster[j] )
    		{
    			mat[i,j] = 0
    			mat[j,i] = 0
    		}
		}
	return (mat)
}

ToIndex <- function(z_mat)
{
	if(is.vector(z_mat))
		return (which.max(z_mat) )
	res <- rep(0, length(z_mat[1,]))
	for(i in 1:length(z_mat[1,]))
		res[i] <- which.max(z_mat[,i])
	return (res)
}

SpectralCluster <- function(S, k) 
{
	S <- as.matrix(S)
	S <- abs(S)
	D <- diag(rowSums(S))
	V <- eigen(solve(D, S))$vectors[, 2:(k + 1)]
	km <- kmeans(V, k)
	return(km$cluster)
}

Cholesky <- function(A)
{
	U <- chol(A)   ## start from factor given by chol 
	D <- diag(U)   ## extract sqrt(D) 
	L <- t(U/D)     ## get unit lower triangular factor 

	L <- L %*% (diag(D))
	return(L )
}

DistCluster <- function(path, true.cluster) {
	p <- nrow(path)
	nsim <- ncol(path)
	for (i in 1:nsim) {
		temp <- path[,i]
		grp <- unique(temp)
		for (j in 1:length(grp)) {
			temp[temp==grp[j]] <- j-1
		}
		path[,i] <- temp
	}
	ncluster <- max(path)+1
	part <- numeric(nsim)
	for (i in 1:nsim) {
		part[i] <- paste(path[,i]+1, collapse=',')
	}
	dist <- table(part)
	name <- names(dist)
	acc <- rep(0, length(name) )
	for(i in 1:length(acc))
		acc[i] <- CompareCluster(true.cluster, name[i])
	return(data.frame(name=names(dist), prob=dist[]/nsim, acc = acc))
}

CheckGroup <- function(z)
{
	K <- max(z)
	for(i in 1:K)
	{
		s <- sum( z == i)
		if(s == 0)
		{
			cat( paste("no group ", s, "\n", sep=""))
			return(F)
		}
	}
}

CheckStationary <- function(z.path)
{
	n <- length(z.path[1,])
	p <- length(z.path[,1])
	station <- -1
	for(i in 2:n)
	{
		count <- sum( z.path[, i-1] == z.path[,i])
		if(count == p)
		{	
			if(station == -1)
				station <- i
		}
		else
		{
			station <- -1
			#print(i)
		}
		cat( paste("count = ", count, "\tp = ", p, "\tstation = ", station, "\n", sep="") )
	}
	return(station)
}


StockData <- function() {
	data(stockdata)
	Y = log(stockdata$data[2:1258, ] / stockdata$data[1:1257, ])
	category <- unique(stockdata$info[, 2])
	label <- numeric(dim(stockdata$data)[2])
	for (i in 1:length(category)) {
		label[stockdata$info[, 2] == category[i]] <- i
	}
	return(list(returns = Y, label = label))
}

Sparsity <- function(mat)
{
	mat.lower <- mat[ lower.tri(mat, diag=F)]
	return( sum(mat.lower==0) / length(mat.lower) )
}

star.glasso <- function(x, subsample=30, beta=0.2)
{
	lambda <- c( seq(0.01, 0.1, by=0.01), seq(0.2, 0.5, by=0.1) )
	n <- nrow(x)
	p <- ncol(x)
	b <- min(as.integer( 10 * sqrt(n) ), 0.8*n)
	#beta <- 0.2
	for(lam in lambda)
	{
		count <- matrix(0, p, p)
		for(i in 1:subsample)
		{
			x.sub.index <- sample(1:n, size=b, replace=F)
			x.sub <- x[x.sub.index, ]
			wi.sub <- glasso( cov(x.sub), lam)$wi
			count <- count + (wi.sub!=0)
		}
		val <- as.vector( count[ lower.tri(count, diag=F)] )
		val <- val / subsample
		val <- val * (1-val) * 2
		beta.est <- mean(val)
		#cat ( paste("l=", lam, "\tbeta=", beta.est, "\tsparsity=", Sparsity(wi.sub), "\n") )
		if( beta.est <= beta)
			break
	}
	return (lam)
}

CRP <- function(p, alpha) {
	z <- rep(0,p)
	z[1] <- 1
	for (i in 2:p) {
		prob <- table(z[1:i-1])
		group <- as.numeric(names(prob))
		prob <- c(prob, alpha)
		prob <- prob/sum(prob)
		group <- c(group, max(group)+1)
		z[i] <- group[sample(group, 1, prob=prob)]
	}
	return(z)
}

SliceSample <- function (x0, g, w=1, m=Inf, lower=-Inf, upper=+Inf, gx0=NULL){
	
	uni.slice.calls <- 0	# Number of calls of the slice sampling function
	uni.slice.evals <- 0	# Number of density evaluations done in these calls
	# Check the validity of the arguments.
	
	if (!is.numeric(x0) || length(x0)!=1
		|| !is.function(g) 
		|| !is.numeric(w) || length(w)!=1 || w<=0 
		|| !is.numeric(m) || !is.infinite(m) && (m<=0 || m>1e9 || floor(m)!=m)
		|| !is.numeric(lower) || length(lower)!=1 || x0<lower
		|| !is.numeric(upper) || length(upper)!=1 || x0>upper
		|| upper<=lower 
		|| !is.null(gx0) && (!is.numeric(gx0) || length(gx0)!=1))
	{ 
		stop ("Invalid slice sampling argument")
	}
	
	# Keep track of the number of calls made to this function.
	
	uni.slice.calls <- uni.slice.calls + 1
	
	# Find the log density at the initial point, if not already known.
	
	if (is.null(gx0)) 
	{ uni.slice.evals <- uni.slice.evals + 1
	  gx0 <- g(x0)
	}
	
	# Determine the slice level, in log terms.
	
	logy <- gx0 - rexp(1)
	
	# Find the initial interval to sample from.
	
	u <- runif(1,0,w)
	L <- x0 - u
	R <- x0 + (w-u)  # should guarantee that x0 is in [L,R], even with roundoff
	
	# Expand the interval until its ends are outside the slice, or until
	# the limit on steps is reached.
	
	if (is.infinite(m))  # no limit on number of steps
	{ 
		repeat
		{ if (L<=lower) break
		  uni.slice.evals <- uni.slice.evals + 1
		  if (g(L)<=logy) break
		  L <- L - w
		}
		
		repeat
		{ if (R>=upper) break
		  uni.slice.evals <- uni.slice.evals + 1
		  if (g(R)<=logy) break
		  R <- R + w
		}
	}
	
	else if (m>1)  # limit on steps, bigger than one
	{ 
		J <- floor(runif(1,0,m))
		K <- (m-1) - J
		
		while (J>0)
		{ if (L<=lower) break
		  uni.slice.evals <- uni.slice.evals + 1
		  if (g(L)<=logy) break
		  L <- L - w
		  J <- J - 1
		}
		
		while (K>0)
		{ if (R>=upper) break
		  uni.slice.evals <- uni.slice.evals + 1
		  if (g(R)<=logy) break
		  R <- R + w
		  K <- K - 1
		}
	}
	
	# Shrink interval to lower and upper bounds.
	
	if (L<lower) 
	{ L <- lower
	}
	if (R>upper)
	{ R <- upper
	}
	
	# Sample from the interval, shrinking it on each rejection.
	
	repeat
	{ 
		x1 <- runif(1,L,R)
		
		uni.slice.evals <- uni.slice.evals + 1
		gx1 <- g(x1)
		
		if (gx1>=logy) break
		
		if (x1>x0) 
		{ R <- x1
		}
		else 
		{ L <- x1
		}
	}
	
	# Return the point sampled, with its log density attached as an attribute.
	
	attr(x1,"log.density") <- gx1
	return (x1)
}

post_zi <- function(i, z, x, omega, df, alpha, omega_hat, shrink, int = F) {
	n <- nrow(x)
	p <- ncol(x)
	z[i] <- 0
	znum <- table(z[-i])[]
	znum <- c(znum, alpha)
	znum <- znum/sum(znum)
	zname <- sort(unique(z[-i]))
	zname <- c(zname, max(zname)+1)
	prob <- rep(0, length(zname))
	u <- matrix(z, p, p)==matrix(z, p, p, byrow=T)
	for (j in 1:length(zname)) {
		k <- zname[j]
		w <- u
		w[i,] <- z==k
		w[,i] <- z==k
		w[i,i] <- T
		w[w==F] <- shrink
		v <- w*omega_hat
		if (int) {
			prob[j] <- n/2*log(det(v))-(n+df)/2*log(det(diag(p)+v%*%t(x)%*%x/df))+log(znum[j])
		}
		else {
			prob[j] <- -df/2*log(det(v))-0.5*sum(diag(solve(v/df,omega)))+log(znum[j])
		}
	}
	prob <- prob-max(prob)
	prob <- exp(prob)
	prob <- prob/sum(prob)
	return(data.frame(group=zname, prob=prob)[])
}

post_omega <- function(z, x, df, omega_hat, shrink) {
	p <- length(z)
	n <- dim(x)[1]
	u <- matrix(z, p, p)==matrix(z, p, p, byrow=T)
	u[u==F] <- shrink
	v <- u*omega_hat/df
	v <- solve(solve(v)+t(x)%*%x)
	return(rWishart(1, n+df, v)[,,1])
}

post_alpha <- function(z, thin=10) {
	p <- length(z)
	k <- length(unique(z))
	alpha <- rgamma(1,1,1)
	for (j in 1:thin) {
		alpha <- SliceSample(alpha, function (x) lgamma(x)-lgamma(x+p)+k*log(x)-x/2, lower=0)
	}
	return(alpha)
}

Gibbs <- function(x, nsim, df, shrink, zstart = NULL, rate.alpha = 2/p, 
				  int = F, display = F, alpha.const=-1, fixed=NULL, omega.start=NULL)
{
	n <- nrow(x)
	p <- ncol(x)
	
	ss <- cov(x)
	min.eig <- min( abs(eigen(ss)$val) )
	if(min.eig < 10^-10)
		omega_hat <- glasso(ss, 0.001)$wi
	else
		omega_hat <- solve(ss)
	
	if(is.null(omega.start) )
		omega <- omega_hat
	else
		omega <- omega.start
	
	if(alpha.const==-1)
		alpha <- rgamma(1, 1, rate.alpha)
	else
		alpha <- alpha.const
	
	if (is.null(zstart)) {
		z <- CRP(p, alpha)
	}
	else z <- zstart
	
	if(display)
		cat(z, "\n")
	zpath <- matrix(0, ncol = nsim, nrow = p)
	zpath[,1] <- z
	apath <- rep(0, nsim)
	apath[1] <- alpha
	for (i in 2:nsim) {
		cease <- 0
		for (j in 1:p) {
			if(!is.null(fixed) && fixed[j])
				next
			
			zpost <- post_zi(j, z, x, omega, df, alpha, omega_hat, shrink, int)
			z[j] <- sample(zpost$group, 1, T, zpost$prob)
			if (max(zpost$prob)>0.99) cease <- cease+1
			if (display) cat(z[j], " ")
		}
		if(display)
			cat("\n")
		zpath[,i] <- z
		if (!int) {
			omega <- post_omega(z, x, df, omega_hat, shrink)
		}
		if(alpha.const == -1)
			alpha <- post_alpha(z)
		apath[i] <- alpha
		if (cease == p) break
	}
	omega <- post_omega(z, x, df, omega_hat, shrink)
	result <- list()
	result$zpath <- zpath
	#result$omega <- omega
	result$apath <- apath
	result$z <- z
	return(result)
}

LogMultiGamma <- function(p, n) {
	g <- p*(p-1)/4 * log(pi)
	for (j in 1:p) {
		g <- g + lgamma(n+(1-j)/2)
	}
	return(g)
}

lmgq <- function(n, p) {
	g <- 0
	for (j in 1:p) {
		g <- g + lgamma((n+p+1-j)/2) - lgamma((p+1-j)/2)
	}
	return(g)
}

post_zi.sun <- function(i, z, x, v, alpha) {
	n <- nrow(x)
	p <- ncol(x)
	
	znum <- table(z[-i])[]
	zname <- sort(unique(z[-i]))
	k <- length(zname)
	prob <- rep(0, k+1)
	
	for (j in 1:k) {
		z[i] <- zname[j]
		pj <- znum
		pj[j] <- pj[j]+1
		for (t in 1:k) {
			ind <- z==zname[t]
			vj <- v[ind,ind]/pj[t]
			xj <- x[,ind]
			prob[j] <- prob[j] + lmgq(n,pj[t]) + n/2*log(det(as.matrix(vj))) -(n+pj[t])/2*log(det(as.matrix(diag(pj[t])+vj%*%t(xj)%*%xj))) 
		}
		prob[j] <- prob[j]+log(znum[j])
	}
	z[i] <- zname[k]+1
	pj <- c(znum,1)
	for (t in 1:(k+1)) {
		ind <- z==c(zname,zname[k]+1)[t]
		vj <- v[ind,ind]/pj[t]
		xj <- x[,ind]
		prob[k+1] <- prob[k+1] + lmgq(n,pj[t]) + n/2*log(det(as.matrix(vj))) - (n+pj[t])/2*log(det(as.matrix(diag(pj[t])+vj%*%t(xj)%*%xj))) 
	}
	prob[k+1] <- prob[k+1]+log(alpha)
	
	prob <- prob-max(prob)
	prob <- exp(prob)
	prob <- prob/sum(prob)
	return(data.frame(group=c(zname,zname[k]+1), prob=prob)[])
}

Gibbs.sun <- function(x, nsim, zstart = NULL, rate.alpha = 2/p){
	n <- nrow(x)
	p <- ncol(x)
	#omega_hat <- solve(cov(x))
	alpha <- rgamma(1, 1, rate.alpha)
	if (is.null(zstart)) {
		z <- CRP(p, alpha)
	}
	else z <- zstart
	
	result <- list()
	result$zpath <- z
	result$apath <- alpha
	for (i in 2:nsim) {
		cat("run ", i, "\n")
		cease <- 0
		for (j in 1:p) {
			zpost <- post_zi.sun(j, z, x, diag(p), alpha)
			z[j] <- sample(zpost$group, 1, T, zpost$prob)
            cat(z[j], " ")
			if (max(zpost$prob)>0.99) cease <- cease+1
		}
        cat("\n")
		result$zpath <- cbind(result$zpath, z)
		alpha <- post_alpha(z)
		result$apath <- cbind(result$apath, alpha)
		if (cease == p) break
	}
	result$z <- z
	return(result)
}


post_zi.sun.I <- function(z, index, alpha, n, p, sigma_hat, prev.info) {
	
	#z <- z.test
	znum <- table(z[-index])[]
	zname <- sort(unique(z[-index]))
	k <- length(zname)
	zname <- c(zname, zname[k]+1)
	znum <- c(znum, alpha)
	prior <- log( znum / sum(znum) )
	log.prob <- rep(0, k+1)
	
	ori.z <- z[index]
	
	for(i in 1:(k+1) )
	{
		z[index] <- zname[i]
		val <- prior[i]
		
		if(is.null(prev.info) )
		{
			for(j in zname)
			{
				ind <- z == j 
				p.new <- sum(ind)
				if(p.new == 0)
					next
			
				v.hat <- sigma_hat[ind, ind]			
			
				v.prime <- (n+p.new)/2 * log( det(diag(p.new) + 1/p.new * v.hat))
				val <- val + lmgq(n, p.new) - n*p.new/2 * log(p.new) - v.prime
				#cat(v.prime, "\t")
			}
		}
		else
		{
			for(j in zname)
			{
				if(j == ori.z || j == zname[i])
				{
					ind <- z == j 
					p.new <- sum(ind)
					if(p.new == 0)
						next
					
					v.hat <- sigma_hat[ind, ind]			
					
					v.prime <- (n+p.new)/2 * log( det(diag(p.new) + 1/p.new * v.hat))
					val <- val + lmgq(n, p.new) - n*p.new/2 * log(p.new) - v.prime
					#cat(v.prime, "\t")
				}
				else
					val <- prev.info[j]
				
			}
		}
		log.prob[i] <- val
		#cat(val, "\n")
	}
	prev.info <- rep(0, max(zname))
	for(j in 1:length(zname) )
		prev.info[zname[j]] <- log.prob[j]
	
	log.prob <- log.prob - max(log.prob)
	prob <- exp(log.prob)
	prob <- prob / sum(prob)
	return(list(group=zname, prob=prob, prev.info=NULL))
}

Gibbs.sun.I <- function(x, nsim, zstart = NULL, rate.alpha = 2/p){
	n <- nrow(x)
	p <- ncol(x)
	sigma_hat <- (n-1) * cov(x)
	
	alpha <- rgamma(1, 1, rate.alpha)
	#alpha <- 1
	
	if (is.null(zstart)) {
		z <- CRP(p, alpha)
	}
	else z <- zstart
	
	result <- list()
	result$zpath <- z
	result$apath <- alpha
	
	prev.info <- NULL
	cat("initializing finished: ", z, "\n")
	for (i in 2:nsim) {
		cat("run ", i, "\n")
		cease <- 0
		for (j in 1:p) {
			zpost <- post_zi.sun.I(z, j, alpha, n, p, sigma_hat, prev.info)
			
			prev.info <- zpost$prev.info
			
			z[j] <- sample(zpost$group, 1, T, zpost$prob)
			cat(z[j], " ")
			if (max(zpost$prob)>0.99) cease <- cease+1
		}
		cat("\n")
		result$zpath <- cbind(result$zpath, z)
		alpha <- post_alpha(z)
		result$apath <- cbind(result$apath, alpha)
		if (cease == p) break
	}
	result$z <- z
	return(result)
}


#test <- function(p, n, prob.in, prob.out, n.cluster, n.run, df.muti)
test <- function(x)
{
	if(F){
		set.seed(1)
		p <- 50
		n <- 200
		prob.in <- 1
		prob.out <- 0
		n.cluster <- 4
		n.run <- 10
	}
	
	
	n <- nrow(x)
	p <- ncol(x)
	n.run <- 10

	z <- 1:p
	#cat( paste("p = ", p, "\tn= ", n,"\n", sep=""))
	max.iteration <- 1
	#max.iteration <- 1
	max.group <- -1
	
	df <- n
	for(iter in 1:max.iteration)
	{
		gibbs.tmp <- Gibbs(x,n.run, df, shrink=0, zstart=1:p)
		#cat("iteration ", iter, "\n")
		#gibbs.tmp <- Gibbs.sun(x,n.run,zstart=1:p)
		n.group <- length( table( gibbs.tmp$z) )
		if(n.group > max.group)
		{
			max.group <- n.group
			gibbs <- gibbs.tmp
		}
	}
	if( length(table(gibbs$z))<= 2)
		cat( "trapped! result may not validate!\n")
	
	return( list(path=gibbs$zpath, z=gibbs$z, z.true=l$cluster) )
}

test.sc <- function(x, n.cluster)
{
	n <- nrow(x)
	p <- ncol(x)
	#lambda.star <- star.glasso(x, 30, 0.05)
	lambda.star <- 0.01
	cat("lambda.star=", lambda.star, "\n")
	sc.glasso <- SpectralCluster( glasso(cov(x), lambda.star)$wi, n.cluster )
	sc.cov <- SpectralCluster(cov(x), n.cluster)
	sc.prec <- SpectralCluster(solve(cov(x)), n.cluster)
	return( list(sc.glasso=sc.glasso, sc.cov=sc.cov, sc.prec=sc.prec))
}
